/**********************************************************************
 * Copyright (c) 2010, 2016 IBM Corporation
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors: 
 * IBM - Initial API and implementation
 **********************************************************************/

package org.eclipse.hyades.trace.ui.internal.launcher;

import java.util.ArrayList;
import java.util.List;

import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IConfigurationElement;
import org.eclipse.core.runtime.Platform;
import org.eclipse.debug.core.ILaunchConfiguration;
import org.eclipse.debug.core.ILaunchConfigurationWorkingCopy;
import org.eclipse.debug.ui.ILaunchConfigurationTab;
import org.eclipse.hyades.trace.ui.IGenericAttachableAgentListExtension;
import org.eclipse.hyades.trace.ui.IProfilerFilterProcessorExtension;
import org.eclipse.hyades.trace.ui.IProfilerTabExtension;
import org.eclipse.hyades.trace.ui.IProfilerUIAdditionExtension;
import org.eclipse.hyades.trace.ui.UIPlugin;
import org.eclipse.hyades.trace.ui.internal.core.TraceProfileFiltersUI;
import org.eclipse.hyades.trace.ui.internal.util.FilterSetElement;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.tptp.platform.common.internal.GenericAttachableAgent;
import org.eclipse.ui.IWorkingSet;

/**
 * Provides support for the extension points that were introduced as part of bug 323330.
 */
public class HyadesTraceUIExtensionSupportUtil {
	
	public enum TraceExtensionUIElementPos { TOP, BOTTOM };
	
	private final static String TEXT_TOP = "top";
	private final static String TEXT_BOTTOM = "bottom";
	private final static String ATTRIBUTE_POSITION = "position";
	
	private static TraceExtensionUIElementPos convertTraceExtensionUIElementPos(String str) {
		if(str == null) return null;
		
		str = str.trim();
		
		if(str.equalsIgnoreCase(TEXT_TOP)) return TraceExtensionUIElementPos.TOP;
		if(str.equalsIgnoreCase(TEXT_BOTTOM)) return TraceExtensionUIElementPos.BOTTOM;

		return null;
		
	}
	
	/** to enable or disable the customize auto-filter button on filterUI page*/
	public static void enableAutoDisplayButton(final TraceProfileFiltersUI TPF){
		IConfigurationElement[] config = Platform.getExtensionRegistry()
				.getConfigurationElementsFor(
						UIPlugin.PLUGIN_ID+".profilingTabUIAddition");

			for (IConfigurationElement e : config) {
				try {
					final Object o = e.createExecutableExtension("class");
					if (o instanceof IProfilerUIAdditionExtension) {
						IProfilerUIAdditionExtension i = (IProfilerUIAdditionExtension)o;
						i.enableAutoDisplayButton(TPF);
					}
				} catch (CoreException e1) {
					e1.printStackTrace();
				}catch(AbstractMethodError e2){
					//if the method is not implemented, then do nothing
					return;
				}
			}
	}
	
	/** add additional button "customize auto-filter* to the filterUI page as the UI support for 
	 * displaying auto-application based filter content.
	 */
	public static void addExtensionUIDisplayButton(Composite composite, TraceProfileFiltersUI TPF) {

		IConfigurationElement[] config = Platform.getExtensionRegistry()
			.getConfigurationElementsFor(
					UIPlugin.PLUGIN_ID+".profilingTabUIAddition");

		for (IConfigurationElement e : config) {
			
			
			try {
				final Object o = e.createExecutableExtension("class");
				if (o instanceof IProfilerUIAdditionExtension) {
					IProfilerUIAdditionExtension i = (IProfilerUIAdditionExtension)o;
					i.addExtensionUIDisplayButton(composite, TPF);
				}
			} catch (CoreException e1) {
				e1.printStackTrace();
			}catch(AbstractMethodError e2){
				//if the method is not implemented, then do nothing
				return;
			}
		}

	}
	
	/** Provides extensions the parent composite of the profiling tab, the current launch configuration,
	 * and a reference to the profile tab from which the composite is coming from. (Bug 323330) */
	public static void callExtensionUIElements(Composite composite, ILaunchConfiguration launchConf, ProfileTab profileTab, TraceExtensionUIElementPos position) {

		IConfigurationElement[] config = Platform.getExtensionRegistry()
			.getConfigurationElementsFor(
					UIPlugin.PLUGIN_ID+".profilingTabUIAddition");

		for (IConfigurationElement e : config) {
			
			String extensionPosition = e.getAttribute(ATTRIBUTE_POSITION);
			TraceExtensionUIElementPos extPosConst = convertTraceExtensionUIElementPos(extensionPosition);

			// Ignore unmatched text
			if(extPosConst == null) continue;

			// Ignore text that doesn't match the current position of the calling method
			if(extPosConst != position) continue;
			
			try {
				final Object o = e.createExecutableExtension("class");
				if (o instanceof IProfilerUIAdditionExtension) {
					IProfilerUIAdditionExtension i = (IProfilerUIAdditionExtension)o;
					i.addExtensionUIElements(composite, launchConf, profileTab);
				}
			} catch (CoreException e1) {
				e1.printStackTrace();
			}
		}

	}

	/** Provides the list of tabs from the particular profiling dialog tab group, and the class
	 * from which the extension is being called. (Bug 323330) */
	@SuppressWarnings("rawtypes")
	public static void addOrInformExtensionTabs(List<ILaunchConfigurationTab> l, Class src) {
		
		IConfigurationElement[] config = Platform.getExtensionRegistry()
			.getConfigurationElementsFor(UIPlugin.PLUGIN_ID+".profilingTabList");
		
		for(IConfigurationElement e : config) {
			try {
				final Object o = e.createExecutableExtension("class");
				if(o instanceof IProfilerTabExtension) {
					IProfilerTabExtension i = (IProfilerTabExtension)o;
					i.updateOrAddTabs(l, src);
				}
			} catch (CoreException e1) {
				e1.printStackTrace();
			}
		}
	}
	
	/**Back-end support for displaying the automatic application based profiling filter */
	public static void displayAutoFilterCotent(ILaunchConfiguration conf,IWorkingSet[] workingSets, FilterSetElement FSE){
		IConfigurationElement[] config = Platform.getExtensionRegistry().getConfigurationElementsFor(UIPlugin.PLUGIN_ID+".profilingFilterProcessor");
		for (IConfigurationElement e : config) {
			try {
				final Object o = e.createExecutableExtension("class");
				
				if (o instanceof IProfilerFilterProcessorExtension) {
					IProfilerFilterProcessorExtension i = (IProfilerFilterProcessorExtension)o;
					i.displayAutoFilterCotent(conf, workingSets, FSE);
				}
			} catch (CoreException e1) {
				e1.printStackTrace();
			}catch(AbstractMethodError e2){
				//if the method is not implemented, then do nothing
				return;
			}
		}
	}
	
	
	/** Provides the extension with the launch configuration and working sets, which contain
	 * the filter and filter settings to be processed. (Bug 323330) */
	public static void applyExtensionFilterProcessing(ILaunchConfigurationWorkingCopy launchConf, IWorkingSet[] wsArr) {

		IConfigurationElement[] config = Platform.getExtensionRegistry().getConfigurationElementsFor(UIPlugin.PLUGIN_ID+".profilingFilterProcessor");

		for (IConfigurationElement e : config) {

			try {
				final Object o = e.createExecutableExtension("class");
				
				if (o instanceof IProfilerFilterProcessorExtension) {
					IProfilerFilterProcessorExtension i = (IProfilerFilterProcessorExtension)o;
					i.processFilter(launchConf, wsArr);
				}
			} catch (CoreException e1) {
				e1.printStackTrace();
			}
		}

	}
	
	/** Given host info, return a list of agents that are found on that server */
	public static List<GenericAttachableAgent> getAvailableGenericAgents(String hostName) {

		IConfigurationElement[] config = Platform.getExtensionRegistry().getConfigurationElementsFor(UIPlugin.PLUGIN_ID+".genericAgentAvailability");

		for (IConfigurationElement e : config) {

			try {
				final Object o = e.createExecutableExtension("class");
				
				if (o instanceof IGenericAttachableAgentListExtension) {
					IGenericAttachableAgentListExtension i = (IGenericAttachableAgentListExtension)o;
					return ((IGenericAttachableAgentListExtension) o).getAvailableAgents(hostName);
				}
			} catch (CoreException e1) {
				e1.printStackTrace();
			}
		}
		
		return new ArrayList<GenericAttachableAgent>();

	}
	/** disable or enable functional buttons on filterUI page based on the setting from
	 * launch configuration */
	public static void applyAutoAffect(
			TraceProfileFiltersUI traceProfileFiltersUI) {
		IConfigurationElement[] config = Platform.getExtensionRegistry()
				.getConfigurationElementsFor(
						UIPlugin.PLUGIN_ID+".profilingTabUIAddition");

			for (IConfigurationElement e : config) {
				
				
				try {
					final Object o = e.createExecutableExtension("class");
					if (o instanceof IProfilerUIAdditionExtension) {
						IProfilerUIAdditionExtension i = (IProfilerUIAdditionExtension)o;
						i.applyAutoAffect(traceProfileFiltersUI);
					}
				} catch (CoreException e1) {
					e1.printStackTrace();
				}catch(AbstractMethodError e2){
					//if the method is not implemented, then do nothing
					return;
				}
			}
	}
	
	/** add additional UI checkbox to the filterUI page as an option to enable automatically
	 * determine filter criteria*/
	public static void addExtensionUIAutoFilter(Composite composite, TraceProfileFiltersUI TPF) {
		IConfigurationElement[] config = Platform.getExtensionRegistry()
				.getConfigurationElementsFor(
						UIPlugin.PLUGIN_ID+".profilingTabUIAddition");
			for (IConfigurationElement e : config) {
				
				
				try {
					final Object o = e.createExecutableExtension("class");
					if (o instanceof IProfilerUIAdditionExtension) {
						IProfilerUIAdditionExtension i = (IProfilerUIAdditionExtension)o;
						i.addExtensionUIAutoFilter(composite, TPF);
					}
				} catch (CoreException e1) {
					e1.printStackTrace();
				}catch(AbstractMethodError e2){
					//if the method is not implemented, then do nothing
					return;
				}
			}

		}
	
	/** Modifying exisitng button on filterUI page with new text label*/
	public static void modifiedButtonLabel(Button btn, String newName) {
		IConfigurationElement[] config = Platform.getExtensionRegistry()
				.getConfigurationElementsFor(
						UIPlugin.PLUGIN_ID+".profilingTabUIAddition");

			for (IConfigurationElement e : config) {
				
				
				try {
					final Object o = e.createExecutableExtension("class");
					if (o instanceof IProfilerUIAdditionExtension) {
						IProfilerUIAdditionExtension i = (IProfilerUIAdditionExtension)o;
						i.modifiedButtonLabel(btn, newName);
					}
				} catch (CoreException e1) {
					e1.printStackTrace();
				}catch(AbstractMethodError e2){
					//if the method is not implemented, then do nothing
					return;
				}
			}
		
	}
	
	/**Save user's choice of automatically determine filter criteria from filterUI
	 * into launchconfiguraiton */
	public static void storeAutoAttribute(
			ILaunchConfigurationWorkingCopy workingCopy,
			TraceProfileFiltersUI filterUI) {
		IConfigurationElement[] config = Platform.getExtensionRegistry()
				.getConfigurationElementsFor(
						UIPlugin.PLUGIN_ID+".profilingTabUIAddition");
			for (IConfigurationElement e : config) {
				
				
				try {
					final Object o = e.createExecutableExtension("class");
					if (o instanceof IProfilerUIAdditionExtension) {
						IProfilerUIAdditionExtension i = (IProfilerUIAdditionExtension)o;
						i.storeAutoAttribute(workingCopy, filterUI);
					}
				} catch (CoreException e1) {
					e1.printStackTrace();
				}catch(AbstractMethodError e2){
					//if the method is not implemented, then do nothing
					return;
				}
			}
		
	}
		
	}


