 /********************************************************************** 
 * Copyright (c) 2005 - 2018 IBM Corporation and others. 
 * All rights reserved.   This program and the accompanying materials 
 * are made available under the terms of the Eclipse Public License v1.0 
 * which accompanies this distribution, and is available at 
 * http://www.eclipse.org/legal/epl-v10.html         
 * $Id: ProbeInstrumenter.cpp,v 1.11 2010-08-04 19:05:34 mreid Exp $ 
 * 
 * Contributors: 
 * IBM - Initial API and implementation 
 **********************************************************************/ 

//==============================================================================
// ProbeInstrumenter.cpp
//
// This is the "main" that drives the BCIEngProbe libary to instrument
// class files on disk.
// 
// USAGE: ProbeInstrumenter engine_script class_file [ class_file ... ]
//
// Instruments the named class files. The output file name for each
// is the original name plus ".bci"
//

#ifdef WIN32
#pragma warning(disable:4786)
#endif

// This is a command line tool, so for MVS use native encoding so that strings
// written to the console can be read
#ifdef MVS
#pragma convert("IBM-1047")
#endif

#include "BCIEngProbeInterface.h"
#include "JVMUtils.h"
#include "CommonDef.h"
#include <string>
#include <stdlib.h>
#include <stdio.h>

// Long-path name support for Windows //////////////////////////////////////////

#ifdef _WIN32

#include <windows.h>

static wchar_t *to_wide( UINT cp, const char *s ) {

	wchar_t *wide = NULL;
	int size = 0;
	
	size = MultiByteToWideChar( cp, MB_ERR_INVALID_CHARS, s, -1, NULL, 0 );
	if( !size )
		return NULL;

	wide = (wchar_t*)malloc( size * sizeof(wchar_t) );
	if( !wide )
		return NULL;

	if( !MultiByteToWideChar( cp, MB_ERR_INVALID_CHARS, s, -1, wide, size) ) {
		free( wide );
		return NULL;
	}

	return wide;

}

int tptp_fclose( FILE* fp ) {
	return fclose( fp );
}

int tptp_feof( FILE *fp ) {
	return feof( fp );
}

int tptp_ferror( FILE *fp ) {
	return ferror( fp );
}

int tptp_fflush( FILE *fp ) {
	return fflush( fp );
}

char *tptp_fgets( char *buf, int max, FILE* fp ) {
	return fgets( buf, max, fp );
}

/**
 *********************************************************
 *
 * @brief
 *    An implementation of fopen that correctly handles 
 *    paths exceeding MAX_PATH (259) characters.
 *
 *********************************************************/
FILE *tptp_fopen( const char *filename, const char *mode ) {

	char cwd[ 32768 ];
	char absolutePath[ 32768 ];
	char path[ 32768 ];
	char *slash, *prevprev, *prev, *start, *end;
	size_t pathLength;
	wchar_t *widePath = NULL, *wideMode = NULL;
	FILE *fp = NULL;

	pathLength = strlen(filename);
	if( pathLength+1 > sizeof(path) )
		return NULL;
	strcpy( path, filename );

	if( !GetCurrentDirectory( sizeof(cwd), cwd ) )
		return NULL;

	// First, convert / to \ (extended length paths may not contain /'s)
	slash = strchr( path, '/' );
	while( slash ) {
		*slash = '\\';
		slash = strchr( slash+1, '/' );
	}

	// Convert to absolute path //////////////////////////////////////////////
	
	// Some clever chap has already given us an extended-length absolute path? Say no more!
	if( pathLength > strlen("\\\\?\\") 
	    && (   0 == memcmp(path, "\\\\?\\", strlen("\\\\?\\")) 
	        || 0 == memcmp(path, "\\\\.\\", strlen("\\\\.\\"))) ) {

		if( pathLength+1 > sizeof(absolutePath) )
			return NULL;
		strcpy( absolutePath, path );

	}
	// Drive qualified ..
	else if( pathLength > 3 && isalpha(path[0]) && ':' == path[1] ) {

		// .. absolute path: X:\dir\...
		if( '\\' == path[2] ) {

			if( pathLength+1 > sizeof(absolutePath) )
				return NULL;

			strcpy( absolutePath, path );

		}
		// .. relative path: X:dir\...
		else {

			char drive[] = "X:";
			char drivewd[ 32768 ];

			drive[0] = path[0];
			SetCurrentDirectory( drive );
			
			if( !GetCurrentDirectory( sizeof(drivewd), drivewd ) )
				return NULL;

			if( strlen(drivewd)+1 + pathLength-2 + 1 > sizeof(absolutePath) )
				return NULL;

			strcpy( absolutePath, drivewd );
			strcat( absolutePath, "\\" );
			strcat( absolutePath, path+2 );

			SetCurrentDirectory( cwd );

		}

	}
	// A UNC: \\server\...
	else if( pathLength > strlen("\\\\") && 0 == memcmp(path, "\\\\", strlen("\\\\")) ) {

		if( strlen("\\\\?\\UNC\\") + pathLength-2 + 1 > sizeof(absolutePath) )
			return NULL;

		strcpy( absolutePath, "\\\\?\\UNC\\" );
		strcat( absolutePath, path+2 );

	}
	// Drive absolute: \dir1\...
	else if( pathLength > 1 && '\\' == path[0] ) {

		if( strlen("X:") + strlen(path) + 1 > sizeof(absolutePath) )
			return NULL;

		strncpy( absolutePath, cwd, strlen("X:") );
		strcpy( absolutePath + strlen("X:"), path );

	}
	// Relative to cwd: dir1\...
	else {

		if( strlen(cwd) + 1 + pathLength + 1 > sizeof(absolutePath) )
			return NULL;

		strcpy( absolutePath, cwd );
		strcat( absolutePath, "\\" );
		strcat( absolutePath, path );

	}

	// De-relative-ize the path (collapse .\ and ..\ segments)
	pathLength = strlen( absolutePath );
	prevprev = NULL;                      // previous of previous segment
	prev  = NULL;                         // previous segment
	
	// start of current segment
	if( 0 == memcmp( "\\\\?\\", absolutePath, strlen("\\\\?\\") )
	 || 0 == memcmp( "\\\\.\\", absolutePath, strlen("\\\\.\\") ) )
		start = strchr( absolutePath + strlen("\\\\?\\") + 1, '\\' );
	else
		start = strchr( absolutePath, '\\' ); 

	if( start )                           // end of current segment
		end = strchr( start+1, '\\' );
	else
		end = NULL;
	while( end ) {

		// Collapse '\..' and move up one directory level
		if( 0 == memcmp( "\\..", start, max((int)strlen("\\.."), end - start) ) ) {

			// If there is no previous segment, we have a mal-formed path.
			if( !prev )
				return NULL;

			memmove( prev, end, (absolutePath + pathLength) - end + 1 );

			end      = strchr( prev+1, '\\' );
			start    = prev;
			prev     = prevprev;
			while( prevprev > absolutePath && '\\' != *(--prevprev) )
				;
			if( prevprev == absolutePath )
				prevprev = NULL;

		// Collapse '.\' or '\\'
		} else if( 0 == memcmp( "\\.", start, max((int)strlen("\\."), end - start) )
		        || 0 == memcmp( "\\\\", start, max((int)strlen("\\\\"), end - start) ) ) {

			memmove( start, end, (absolutePath + pathLength) - end + 1 );
			end   = strchr( start+1, '\\' );

		// Advance to the next path segment
		} else {

			prevprev = prev;
			prev     = start;
			start    = end;
			end      = strchr( start+1, '\\' );

		}

	}

	// Prepend the extended-length path prefix if necessary
	if( 0 != memcmp(absolutePath, "\\\\?\\", strlen("\\\\?\\"))
	 && 0 != memcmp(absolutePath, "\\\\.\\", strlen("\\\\.\\")) ) {

		if( strlen(absolutePath) + strlen("\\\\?\\") + 1 > sizeof(path) )
			return NULL;

		strcpy( path, "\\\\?\\" );
		strcat( path, absolutePath );

	}
	else
		strcpy( path, absolutePath );

	// Finally, convert to wchar_t ///////////////////////////////////////////

	widePath = to_wide( CP_ACP, path );
	if( !widePath )
		goto done;

	wideMode = to_wide( CP_ACP, mode );
	if( !wideMode )
		goto done;

	fp = _wfopen( widePath, wideMode );

done:
	if( widePath )
		free( widePath );
	if( wideMode )
		free( wideMode );

	return fp;

}

int tptp_fprintf( FILE *fp, const char *fmt, ... ) {

	int ret;

	va_list argv;

	va_start( argv, fmt );
	ret = vfprintf( fp, fmt, argv );
	va_end( argv );

	return ret;

}

size_t tptp_fread( void *dst, size_t el, size_t count, FILE *fp ) {
	return fread( dst, el, count, fp );
}

void tptp_frewind( FILE *fp ) {
	rewind( fp );
}

int tptp_fseek( FILE *fp, long ofs, int origin ) {
	return fseek( fp, ofs, origin );
}

long tptp_ftell( FILE *fp ) {
	return ftell( fp );
}

size_t tptp_fwrite( const void *src, size_t el, size_t count, FILE *fp ) {
	return fwrite( src, el, count, fp );
}

#else // other platforms are fine as-is

#define tptp_fclose fclose
#define tptp_feof feof
#define tptp_ferror ferror
#define tptp_fflush fflush
#define tptp_fgets fgets
#define tptp_fopen fopen
#define tptp_fprintf fprintf
#define tptp_fread fread
#define tptp_frewind rewind
#define tptp_fseek fseek
#define tptp_ftell ftell
#define tptp_fwrite fwrite

#endif

USE_NAMESPACE(std);

//------------------------------------------------------------------------------
class CBCIEngProbeInstrumenter
{
public:
	CBCIEngProbeInstrumenter();
	~CBCIEngProbeInstrumenter();
	void InstrumentClassFile(CSTR i_szName);
	void ParseProbeDescriptors(CSTR i_szName);
	unsigned int InterpretArgs(const char* szArgs);
	void EnableJava6Verifier(bool i_enableVerifier);

private:
	CBCIEngInterface*	m_peng;
	bool m_enable_java6_verifier;
};


//------------------------------------------------------------------------------
CBCIEngProbeInstrumenter::CBCIEngProbeInstrumenter()
{
	CreateBCIEngine((void**)&m_peng);
	m_peng->SetAllocator(malloc);
	m_enable_java6_verifier = false;
}

//------------------------------------------------------------------------------
CBCIEngProbeInstrumenter::~CBCIEngProbeInstrumenter()
{
	DestroyBCIEngine(m_peng);
}

enum {
	LOG_LEVEL_FINEST = 1,
	LOG_LEVEL_FINER,
	LOG_LEVEL_FINE,
	LOG_LEVEL_CONFIG,
	LOG_LEVEL_INFO,
	LOG_LEVEL_WARNING,
	LOG_LEVEL_SEVERE,
	LOG_LEVEL_NONE
};

const char* logLevelStrings[] = {
	"(zero)",
	"finest",
	"finer",
	"fine",
	"config",
	"info",
	"warning",
	"severe",
	"none",
	NULL
};

static int logLevel = LOG_LEVEL_SEVERE;
JNIEnv *env = NULL;
JavaVM *jvm = NULL;

//------------------------------------------------------------------------------
int 
callback(CSTR i_clName, size_t len, unsigned flags)
{
	switch (flags) {
		case BCIENGINTERFACE_CALLBACK_MODULE:
			// Query: do you want to permit instrumentation of this module? (false will veto)
			return true;
		case BCIENGINTERFACE_CALLBACK_METHOD:
			// Query: false will veto instr of this method (false will veto)
			return true;
		case BCIENGINTERFACE_CALLBACK_MODULE_INSTR:
			// Notice: I instrumented this module, really made changes in it
			if (logLevel <= LOG_LEVEL_FINE) {
				tptp_fprintf(stderr, "[Probekit: class %s has been instrumented]\n", i_clName);
			}
			break;
		case BCIENGINTERFACE_CALLBACK_METHOD_INSTR:
			// Don't log individual methods that are instrumented
			break;
		default:
			// We don't know of any other messages
			break;
	}
	return true;
}

//------------------------------------------------------------------------------
//void *CBCIEngProbeInstrumenter::GetJNIEnv(void** ptr)
void *GetJNIEnv(void** ptr)
{
	return (void *)env;
}

//------------------------------------------------------------------------------
void 
CBCIEngProbeInstrumenter::ParseProbeDescriptors(CSTR i_szName)
{
	char* pchProbe = NULL;
	size_t cbProbe = 0;
	FILE* pfProbe = tptp_fopen(i_szName, "r");

	if (NULL == pfProbe)
		throw "Can't open probe descriptors file";
	tptp_fseek(pfProbe, 0, SEEK_END);
	cbProbe = tptp_ftell(pfProbe) + 2;
	pchProbe = (char*)malloc(cbProbe + 1);
	tptp_fseek(pfProbe, 0, SEEK_SET);
	cbProbe = tptp_fread(pchProbe, 1, cbProbe, pfProbe);
	pchProbe[cbProbe] = '\0';

	if (logLevel <= LOG_LEVEL_FINER) {
		// Dump the script file we got
		// Dump every line separately
		// Remember that lines might be null-separated.
		// (We made sure the whole thing is null terminated.)
		tptp_fprintf(stderr, "[Probekit: start of script file]\n");
		tptp_fprintf(stderr, "%s\n", pchProbe);
		tptp_fprintf(stderr, "[Probekit: end of script file]\n");
	}

	// Initialize the BCI engine with this script file
	m_peng->Initialize(pchProbe, cbProbe);

	// Request all callbacks
	m_peng->SetCallback(callback, -1);
	free(pchProbe);
	tptp_fclose(pfProbe);
}

//------------------------------------------------------------------------------
void 
CBCIEngProbeInstrumenter::InstrumentClassFile(CSTR i_szName)
{
	string strOutName = i_szName;
	size_t cbIn, cbOut;
	void *pIn, *pOut;

	if (logLevel <= LOG_LEVEL_FINEST) {
		tptp_fprintf(stderr, "[Probekit processing class file %s]\n", i_szName);
	}

	// Open and read class file
	FILE* pfIn = tptp_fopen(i_szName, "rb");
	if (NULL == pfIn)
		throw "Can't open class file";
	tptp_fseek(pfIn, 0, SEEK_END);
	cbIn = tptp_ftell(pfIn);
	pIn = malloc(cbIn);
	tptp_fseek(pfIn, 0, SEEK_SET);
	tptp_fread(pIn, 1, cbIn, pfIn);
	tptp_fclose(pfIn);

	// Instrument
	if(m_enable_java6_verifier) {
		m_peng->Instrument2(nullptr, pIn, cbIn, &pOut, &cbOut, GetJNIEnv);
	}
	else {
		m_peng->Instrument(nullptr, pIn, cbIn, &pOut, &cbOut);
	}
	

	// Write the output. Notice that maybe pOut == pIn
	// if the instrumentation engine decided not to do anything.
	// TODO: flag this somehow to our caller, who might care.
	// TODO: let the caller send us a flag saying not to bother writing *.class.bci in this case.

	if (logLevel <= LOG_LEVEL_FINEST) {
		if (pOut == pIn) {
			tptp_fprintf(stderr, "[Probekit did not change this class file]\n");
		}
	}

	strOutName += ".bci";
	FILE* pfOut = tptp_fopen(strOutName.c_str(), "wb");
	tptp_fwrite(pOut, 1, cbOut, pfOut);
	tptp_fclose(pfOut);

	free(pIn);	
	if (pOut != pIn) {
		// Only free pOut if it's not the same as pIn (which we just freed).
		// It will be the same if the instrumenation engine decided not to do anything.
		free(pOut);
	}
}

//------------------------------------------------------------------------------
void CBCIEngProbeInstrumenter::EnableJava6Verifier(bool i_enableVerifier)
{
	m_peng->UseStackMapRecalculation(i_enableVerifier);
	m_enable_java6_verifier = i_enableVerifier;
	
}
//------------------------------------------------------------------------------
static void PrintHelp()
{
	tptp_fprintf(stderr, "Usage: ProbeInstrumenter [ opts ] engine-script-file class-file [ class-file ... ]\n");
	tptp_fprintf(stderr, "Options are:\n");
	tptp_fprintf(stderr, "\t-loglevel xxx\t(xxx = none, little, some, most, all)\n");
}


//------------------------------------------------------------------------------
//
// Subroutines to turn error codes in to messages.
//
// The getErrorMessage function returns a string which the caller must free!
//
const char* source_strings[] = {
		"(unknown)",
		"driver-to-engine interface",
		"class file manager",
		"Generic module manager",
		"Java module manager",
		"Generic insertion engine",
		"Java insertion engine",
		"Probekit insertion engine",
		"Driver",
		NULL
	};

struct MessageCatalogEntry {
  int src;			// src == -1 means "end of list"
  int code;
  const char* msg;
};

MessageCatalogEntry message_catalog[] = {
	{ 1, 1, "unspecified error" }, 
	{ 1, 2, "bad arguments in probescript" }, 
	{ 1, 3, "bad filter type in probescript" }, 
	{ 2, 1, "invalid method called (internal error)" }, 
	{ 2, 2, "an internal error occurred" }, 
	{ 2, 3, "class file has a bad magic number" }, 
	{ 2, 4, "class file has a bad version" }, 
	{ 2, 5, "class file has a bad entry in the constant pool" }, 
	{ 2, 6, "class file has a bad field reference" }, 
	{ 2, 7, "class file has a bad interface reference" }, 
	{ 2, 8, "class file has a bad method reference" }, 
	{ 2, 9, "class file has a bad value for a constant" }, 
	{ 2, 10, "internal error regarding constant pool type management" }, 
	{ 2, 11, "index out of bounds" }, 
	{ 3, 1, "unknown error" }, 
	{ 3, 2, "internal error" }, 
	{ 3, 3, "invalid module" }, 
	{ 3, 4, "parse error" }, 
	{ 3, 5, "invalid method called (internal error)" }, 
	{ 3, 6, "ouput code size too large (code overrun)" }, 
	{ 5, 1, "unknown error" }, 
	{ 5, 2, "invalid opcode encountered" }, 
	{ 5, 3, "class/module is already instrumented" }, 
	{ 5, 4, "can not instrument this class/module" }, 
	{ 5, 5, "internal error" }, 
	{ 7, 1, "internal error" }, 
	{ 8, 1, "Error loading the jvm library. Put the jre/bin and jre/bin/classic in the path." }, 
	{ 8, 2, "Error resolving function from the jvm library" }, 
	{ 8, 3, "Error instantiating the jvm" }, 
	{ -1, -1, NULL }, 
};

// IMPORTANT!
// The getErrorMessage function returns a string which the caller must free!

char* getErrorMessage(int code, int src) 
{
	if (src < 9) {
		const char* srcmsg = source_strings[src];
		for (int i = 0; message_catalog[i].src != -1; i++) {
			MessageCatalogEntry& mce = message_catalog[i];
			if (mce.src == src && mce.code == code) {
				const char* errmsg = mce.msg;
				const char *fmt = "source: %s, message: %s";
				char* returnmsg = (char*)malloc(strlen(fmt) + strlen(srcmsg) + strlen(mce.msg) + 1);
				sprintf(returnmsg, fmt, srcmsg, mce.msg);
				return returnmsg;
			}
		}
		
		// Known source, unknown error code
		const char* fmt = "source: %s, error code: %d";
		char* returnmsg = (char*)malloc(strlen(fmt) + strlen(srcmsg) + 15);
		sprintf(returnmsg, fmt, srcmsg, code);
		return returnmsg;
	}
	
	// Unknown source
	const char* fmt = "source location: %d, error code: %d";
	char *returnmsg = (char*)malloc(strlen(fmt) + 15 + 15);
	sprintf(returnmsg, fmt, src, code);
	return returnmsg;
}

//------------------------------------------------------------------------------
int main(int argc, char* argv[])
{
	tptp_fprintf(stderr, "  --> Beginning of main method\n");
	
	int errorStatus = 0;
	
	bool enableJava6Verification = false;

	int argScan = 1;

	tptp_fprintf(stderr, "  --> Before Check if argScan >= argc\n");
	
	if (argScan >= argc) {
		PrintHelp();
		return 2;
	}
	
	tptp_fprintf(stderr, "  --> Before check -loglevel\n");
	
	if (strcmp(argv[argScan], "-loglevel") == 0) {
		// parse the next arg as the log level
		++argScan;
		if (argScan >= argc) {
			PrintHelp();
			return 2;
		}
		int match;
		for (match = 0; logLevelStrings[match] != NULL; match++ ){
			if (stricmp(argv[argScan], logLevelStrings[match]) == 0) {
				logLevel = match;
				break;
			}
		}
		if (logLevelStrings[match] == NULL) {
			PrintHelp();
			return 2;
		}
		++argScan;
	}

	tptp_fprintf(stderr, "  --> Before check -usejava6verifier\n");

	if (strcmp(argv[argScan], "-usejava6verifier") == 0) {
		enableJava6Verification = true;
		tptp_fprintf(stderr, "  --> enableJava6Verification is TRUE \n");
		++argScan;
	}

    tptp_fprintf(stderr, "  --> Before CBCIEngProbeInstrumenter eng\n");
    
	CBCIEngProbeInstrumenter eng;

	try {
	    tptp_fprintf(stderr, "  --> Before ParseProbeDescriptors\n");
		eng.ParseProbeDescriptors(argv[argScan]);
	}
	catch (...) {
		tptp_fprintf(stderr, "Problem creating instrumentation engine using engine script %s\n", argv[argScan]);
		return 1;
	}

	++argScan;
    tptp_fprintf(stderr, "  --> Before EnableJava6Verifier\n");
	eng.EnableJava6Verifier(enableJava6Verification);

    tptp_fprintf(stderr, "  --> Before Check enableJava6Verification\n");
	if(enableJava6Verification)
	{
	    tptp_fprintf(stderr, "  --> Before Check createJVM\n");
		int status;
		if(status = createJVM(jvm, &env)) {
			char* errorMessage = getErrorMessage(status , 8);	// we must free this string
			tptp_fprintf(stderr, "Probekit error:\n%s", errorMessage);
			free(errorMessage);
			return 2;
		}


	}

    tptp_fprintf(stderr, "  --> Before InstrumentClassFile\n");
    
	// Operate on each input file.
	// A failure in one doesn't stop us operating on others.
	for (int i = argScan; i < argc; i++) {
		try
		{
			eng.InstrumentClassFile(argv[i]);
		}
		catch (CBCIEngProbeInterfaceException& ex)
		{
			unsigned src=ex.GetSource(); 
			unsigned code=ex.GetReason();
			int msgBufSize = 1024;
			char* msgBuf = (char*)malloc(msgBufSize);
			ex.FormatMessage(msgBuf, msgBufSize);
			char* errorMessage = getErrorMessage(code, src);	// we must free this string
			tptp_fprintf(stderr, "Probekit error while processing %s:\n%s\n%s\n", argv[i], errorMessage, msgBuf);
			free(errorMessage);
			free(msgBuf);
			errorStatus = 2;
		}
		catch (char* szMsg)
		{
			tptp_fprintf(stderr, "Probe test exception: %s\n", szMsg);
			errorStatus = 2;
		}
	}
	
	tptp_fprintf(stderr, "  --> Before check if jvm is null\n");
	if(jvm != NULL) 
	{
		jvm->DestroyJavaVM();
	}
	return errorStatus;
}

#ifdef MVS
#pragma convert(pop)
#endif

//= End of ProbeInstrumenter.cpp =======================================================
