/*****************************************************************************
 * Copyright (c) 2012, 2014 IBM Corporation
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *    IBM Corporation - Initial API and implementation
 *
 * $Id$ 
 *****************************************************************************/
#ifndef __OSA_ATOMIC_H__
#define __OSA_ATOMIC_H__

#include "OSA.h"

//
// Inline atomic integer operations.
//
// This header contains platform and architecture-specific implementations for
// some simple atomic operations on integers.
//
// There is no guarantee that atomicInt_t is a primitive type on all platforms,
// so all access to the value of an atomicInt_t must go through one of the 
// functions defined here.
//
// Supported operations:
//
//  int atomicInc( atomicInt_t* val ); - increments `val' and returns the new value.
//  int atomicDec( atomicInt_t* val ); - decrements `val' and returns the new value.
//

#if defined(__linux__) 

#	if defined( IA32_ARCH ) || defined( EM64T_ARCH )

typedef int atomicInt_t;

static inline int atomicInc( atomicInt_t* val ) {

	atomicInt_t after;

	__asm__ __volatile__(
		
		" lock incl (%1)\n"
		" movl (%1), %0"

		: "=a"(after), "+r"(val) : : "memory" );

	return after;
	
}

static inline int atomicDec( atomicInt_t* val ) {
	
	atomicInt_t after;

	__asm__ __volatile__(

		" lock decl (%1)\n"
		" movl (%1), %0"

		: "=a"(after), "+r"(val) : : "memory" );

	return after;

}

#	elif defined( S390_ARCH ) || defined( S390X_ARCH )

typedef int atomicInt_t;

static inline int atomicInc( atomicInt_t* val ) {

	atomicInt_t after;

	__asm__ __volatile__(

	"         l  %%r1,%1    \n"
	"0:       lr %0,%%r1    \n"
	"         ahi %0,1      \n"
	"         cs %%r1,%0,%1 \n"
	"         jl 0b         \n"

	: "=&d"(after), "+m"(*val) : : "r1", "cc", "memory" );

	return after;
	
}

static inline int atomicDec( atomicInt_t* val ) {
	
	atomicInt_t after;

	__asm__ __volatile__(

	"         l  %%r1,%1    \n"
	"0:       lr %0,%%r1    \n"
	"         ahi %0,-1     \n"
	"         cs %%r1,%0,%1 \n"
	"         jl 0b         \n"

	: "=&d"(after), "+m"(*val) : : "r1", "cc", "memory" );

	return after;
	
}

/* The followsing atomic assembly codes are copied from system header file: */
/*   /usr/src/linux-headers-3.13.0-27/arch/powerpc/include/asm/atomic.h */
#elif defined(__linux__) && defined(__PPC64__) && defined(__LITTLE_ENDIAN__)

#define __stringify_in_c(...)	#__VA_ARGS__
#define stringify_in_c(...)	__stringify_in_c(__VA_ARGS__) " "
#define PPC405_ERR77(ra,rb)	stringify_in_c(dcbt	ra, rb;)

#define LWSYNC lwsync
#define PPC_ATOMIC_ENTRY_BARRIER "\n" stringify_in_c(LWSYNC) "\n"
#define PPC_ATOMIC_EXIT_BARRIER  "\n" stringify_in_c(sync) "\n"

typedef int atomicInt_t;

static inline int atomicInc( volatile atomicInt_t* val ) {
	int t;

	__asm__ __volatile__(
	PPC_ATOMIC_ENTRY_BARRIER
"1: lwarx   %0,0,%1         # atomic_inc_return\n\
	addic   %0,%0,1\n"
	PPC405_ERR77(0,%1)
"   stwcx.  %0,0,%1 \n\
	bne-    1b"
	PPC_ATOMIC_EXIT_BARRIER
	: "=&r" (t)
	: "r" (val)
	: "cc", "xer", "memory");

	return t;
}

static inline int atomicDec( volatile atomicInt_t* val ) {
	int t;

	__asm__ __volatile__(
	PPC_ATOMIC_ENTRY_BARRIER
"1: lwarx   %0,0,%1         # atomic_dec_return\n\
	addic   %0,%0,-1\n"
	PPC405_ERR77(0,%1)
"   stwcx.  %0,0,%1\n\
	bne-    1b"
	PPC_ATOMIC_EXIT_BARRIER
	: "=&r" (t)
	: "r" (val)
	: "cc", "xer", "memory");

	return t;
}

#	endif

#elif defined(_AIX)

#include <builtins.h>
typedef int atomicInt_t;

static inline int atomicInc( volatile atomicInt_t* val ) {
	
	atomicInt_t oldval = *val;
	while( 1 ) {

		atomicInt_t newval = oldval + 1;
		if( __compare_and_swap( val, &oldval, newval ) )
			return newval;

	}

}

static inline int atomicDec( volatile atomicInt_t* val ) {
	
	atomicInt_t oldval = *val;
	while( 1 ) {
		
		atomicInt_t newval = oldval - 1;
		if( __compare_and_swap( val, &oldval, newval ) )
			return newval;
		
	}

}

#elif defined(_SOLARIS)

typedef int atomicInt_t;

#	ifdef __cplusplus
#		define OSA_ATOMIC_EXPORT extern "C" OSA_EXPORT
#	else
#		define OSA_ATOMIC_EXPORT OSA_EXPORT
#	endif

OSA_ATOMIC_EXPORT int atomicCas( atomicInt_t* val, int oldval, int newval );

static inline int atomicInc( atomicInt_t* val ) {

	while( 1 ) {

		atomicInt_t oldval = *val;
		atomicInt_t newval = oldval + 1;

		if( 0 == atomicCas( val, oldval, newval ) )
			return newval;

	}

}

static inline int atomicDec( atomicInt_t* val ) {

	while( 1 ) {

		atomicInt_t oldval = *val;
		atomicInt_t newval = oldval - 1;

		if( 0 == atomicCas( val, oldval, newval ) )
			return newval;

	}

}

#elif defined( _SOLARISX86 )

typedef int atomicInt_t;

#	ifdef __cplusplus
#		define OSA_ATOMIC_EXPORT extern "C" OSA_EXPORT
#	else
#		define OSA_ATOMIC_EXPORT OSA_EXPORT
#	endif

OSA_ATOMIC_EXPORT int atomicInc( atomicInt_t* val );
OSA_ATOMIC_EXPORT int atomicDec( atomicInt_t* val );

#elif defined( MVS )

#include <stdlib.h>

typedef cs_t atomicInt_t;

static inline int atomicInc( atomicInt_t* val ) {

	while( 1 ) {

		atomicInt_t oldval = *val;
		atomicInt_t newval = oldval + 1;

		if( 0 == cs( &oldval, val, newval ) )
			return newval;

	}

}

static inline int atomicDec( atomicInt_t* val ) {

	while( 1 ) {

		atomicInt_t oldval = *val;
		atomicInt_t newval = oldval - 1;

		if( 0 == cs( &oldval, val, newval ) )
			return newval;

	}

}

#elif defined( _WIN32 )

// Reduce name clash potential. We only want the def's for LONG and Interlocked(In|De)crement
#define NOGDICAPMASKS
#define NOVIRTUALKEYCODES
#define NOWINMESSAGES
#define NOWINSTYLES
#define NOSYSMETRICS
#define NOMENUS
#define NOICONS
#define NOKEYSTATES
#define NOSYSCOMMANDS
#define NORASTEROPS
#define NOSHOWWINDOW
#define OEMRESOURCE
#define NOATOM
#define NOCLIPBOARD
#define NOCOLOR
#define NOCTLMGR
#define NODRAWTEXT
#define NOGDI
#define NOKERNEL
#define NOUSER
#define NONLS
#define NOMB
#define NOMEMMGR
#define NOMETAFILE
#define NOMINMAX
#define NOMSG
#define NOOPENFILE
#define NOSCROLL
#define NOSERVICE
#define NOSOUND
#define NOTEXTMETRIC
#define NOWH
#define NOWINOFFSETS
#define NOCOMM
#define NOKANJI
#define NOHELP
#define NOPROFILER
#define NODEFERWINDOWPOS
#define NOMCX
#include <windows.h>

typedef LONG atomicInt_t;

static inline int atomicInc( atomicInt_t volatile * val ) {
	
	return InterlockedIncrement( val );
	
}

static inline int atomicDec( atomicInt_t volatile * val ) {
	
	return InterlockedDecrement( val );
	
}

#else
#	error unknown/unsupported platform
#endif

#endif
